/*
 * Copyright (c) 2000, Jan Rkorajski <baggins@pld-linux.org> All rights reserved.
 *
 * pam_utmp 
 *      Update /var/run/utmpx on open/close session
 *
 * 0.1 - initial release
 *
 */

#include "config.h"

#include <errno.h>
#include <fcntl.h>
#include <pty.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <utmp.h>
#include <utmpx.h>
#include <unistd.h>

#define PAM_SM_SESSION

#include <security/pam_modules.h>
#include <security/_pam_macros.h>

/* logging */
/* closelog first to make sure it'll _really_ go to syslog */
static void _log_err(int err, const char *format,...)
{
	va_list args;

	va_start(args, format);
	closelog();
	openlog("pam_utmp", LOG_CONS | LOG_PID, LOG_AUTH);
	vsyslog(err, format, args);
	va_end(args);
	closelog();
}

int update_utmp(int add, const char *user, const char *device, const char *host)
{
    struct utmpx utx;
    struct passwd * pw;
    char * id;
 
    memset(&utx, 0, sizeof(utx));
    if (add)
	utx.ut_type = USER_PROCESS;
    else
	utx.ut_type = DEAD_PROCESS;

    utx.ut_pid = getppid();
    strncpy(utx.ut_line, device, sizeof(utx.ut_line));

    pw = getpwnam(user);
    if (!pw) {
	_log_err(LOG_ALERT, "cannot find user %s\n", user);
	return PAM_CRED_INSUFFICIENT;
    }

    strncpy(utx.ut_user, pw->pw_name, sizeof(utx.ut_user));

    if (host) 
	strncpy(utx.ut_host, host, sizeof(utx.ut_host));

    if (!strncmp("pts/", utx.ut_line, 4)) {
	id = utx.ut_line + 3;
	if (strlen(id) > 4) id++;
    } else {
	id = utx.ut_line + 2;
    }

    strncpy(utx.ut_id, id, sizeof(utx.ut_id));

    gettimeofday(&utx.ut_tv, NULL);

    pututxline(&utx);
    updwtmpx(_PATH_WTMP, &utx);

    return PAM_SUCCESS;
}

/* now the session stuff */
PAM_EXTERN int pam_sm_open_session(pam_handle_t * pamh, int flags,
				   int argc, const char **argv)
{
	int retval;
	const char *user, *dev, *host;

	D(("called"));
	D(("pid: %d, uid: %d, euid: %d", getpid(), getuid(), geteuid()));

	retval = pam_get_item(pamh, PAM_USER, (const void **)&user);
	if (retval != PAM_SUCCESS || user == NULL || *user == '\0') {
		_log_err(LOG_NOTICE, "user unknown");
		return PAM_USER_UNKNOWN;
	}

	retval = pam_get_item(pamh, PAM_TTY, (const void **)&dev);
	if (retval == PAM_SUCCESS) {
		retval = pam_get_item(pamh, PAM_RHOST, (const void **)&host);
		if (retval == PAM_SUCCESS) {
			D(("user: %s, host: %s, dev: %s", user, host, dev));
			retval = update_utmp(1, user, dev, host);
		}
	}

	return retval;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t * pamh, int flags,
				    int argc, const char **argv)
{
	int retval;
	const char *user, *dev, *host;

	D(("called"));
	D(("pid: %d, uid: %d, euid: %d", getpid(), getuid(), geteuid()));

	retval = pam_get_item(pamh, PAM_USER, (const void **)&user);
	if (retval != PAM_SUCCESS || user == NULL || *user == '\0') {
		_log_err(LOG_NOTICE, "user unknown");
		return PAM_USER_UNKNOWN;
	}

	retval = pam_get_item(pamh, PAM_TTY, (const void **)&dev);
	if (retval != PAM_SUCCESS) {
		retval = pam_get_item(pamh, PAM_RHOST, (const void **)&host);
		if (retval == PAM_SUCCESS) {
			D(("host: %s, dev: %s", host, dev));
			retval = update_utmp(0, user, dev, host);
		}
	}

	return retval;
}

/* static module data */

#ifdef PAM_STATIC
struct pam_module _pam_utmp_modstruct =
{
	"pam_utmp",
	NULL,
	NULL,
	NULL,
	pam_sm_open_session,
	pam_sm_close_session,
	NULL
};
#endif
